/*
 *	init.c
 *
 *	initialization routines
 *
 *	$Id: init.c,v 1.71 2012/03/07 16:28:33 conrads Exp $
 *
 *****************************************************************************/

#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "mkreadmes.h"

static	char	templates_dir[PATH_MAX];	/* directory containing the README.*
												template files
											 */
char	categories[MAXCATLIST];				/* list of categories found in the
												top-level Makefile
											 */
int		num_categories = 0;					/* number of categories found */

extern	char	ports_dir[PATH_MAX];		/* top-level ports directory */
extern	char	ports_dir_plus[PATH_MAX];	/* ports_dir + trailing slash */

extern	int		normal_paths;		/* make relative paths relative to current
										working directory instead of ports_dir
									 */
extern	int		quiet;				/* silence normal output */
extern	int		verbose;			/* increase the amount of output */

/* pointers to dynamically allocated buffers for the template files */
extern	char	*template_buf_top;
extern	char	*template_buf_category;
extern	char	*template_buf_port;

/* prototypes for routines in other source files */

extern	int		load_index(void);	/* index.c */

/*****************************************************************************/

static void cleanup(void)
/*
	final cleanup routine, to release all of the memory we allocated

	registered with atexit(3) during initialization, to be called
	automatically when program exits
*/
{
	extern char **index_categories;
	extern char  *index_buf;

	if (index_buf)
		free(index_buf);

	if (template_buf_top)
		free(template_buf_top);

	if (template_buf_category)
		free(template_buf_category);

	if (template_buf_port)
		free(template_buf_port);

	if (index_categories)
		free(index_categories);
}

/*****************************************************************************/

char *skip_leading(const char *str, const char *pattern)
/*
	skip over the string 'pattern' at the beginning of 'str'

	returns	a pointer to the location in 'str' immediately following 'pattern',
	or 'str' if it does not begin with 'pattern'
 */
{
	int len = strlen(pattern);

	return (strncmp(str, pattern, len) == 0) ?
		(char *) (str + len) : (char *) str;	/* using casts here to quiet a
													compiler warning */
}

/*****************************************************************************/

static int gen_categories_list(void)
/*
 *	generate a list of all ports categories from the top-level Makefile
 *
 *	returns number of categories found, or -1 on error
 */
{
	char	Makefile_path[PATH_MAX];
	char	Makefile_buf[LINE_MAX];
	FILE   *Makefile;

	char   *categories_ptr = categories;	/* pointer into categories list */
	char   *category;		/* pointer to category in current input buffer */

	/* open the top-level Makefile for reading */
	strcpy(Makefile_path, ports_dir);
	strcat(Makefile_path, MAKEFILENAME);

	if ((Makefile = fopen(Makefile_path, "r")) == NULL)
	{
		perror(Makefile_path);
		return -1;
	}

	/* scan through all the "SUBDIR += " lines to get all the categories */
	while (fgets(Makefile_buf, LINE_MAX, Makefile) != NULL)
	{
		if ((category = skip_leading(Makefile_buf, "SUBDIR += ")) !=
				Makefile_buf)	/* found a SUBDIR line */
		{
			/* don't want the newline */
			category[strlen(category) - 1] = '\0';

			/* copy category to the categories list and bump pointer to next
				free slot
			 */
			categories_ptr = stpcpy(categories_ptr, category) + 1;
			++num_categories;
		}
	}

	/* add an extra nul character to mark the end of the categories array */
	*categories_ptr = '\0';

	fclose(Makefile);
	return 0;
}

/*****************************************************************************/

char *load_file(const char *path)
/*
	load the file named by 'path' into a dynamically allocated buffer

	buffer will have an extra nul character appended to it to mark end-of-file

	returns a pointer to the buffer on success, or NULL on failure
 */
{
	char   *buffer;		/* pointer to memory obtained via malloc(3) */
	int		fd;			/* file descriptor for open() and close() */
	long	size;		/* file's size, as returned by stat(2) */
	struct	stat sb;	/* will hold the results from stat(2) */

	/* get the file's size */
	if (stat(path, &sb) == -1)
	{
		perror(path);
		return NULL;
	}

	size = sb.st_size;

	/* allocate (size + 1) bytes, so we can mark end of file
		with an extra nul character
	 */
	if ((buffer = malloc(size + 1)) == NULL)
	{
		perror("malloc failure");
		return NULL;
	}

	/* open and read the entire file into memory */
	if ((fd = open(path, O_RDONLY|O_DIRECT)) == -1)
	{
		perror(path);
		return NULL;
	}

	if (read(fd, buffer, size) != size)
	{
		perror(path);
		close(fd);
		return NULL;
	}

	close(fd);

	/* add an extra nul to mark end of buffer */
	buffer[size] = '\0';

	return buffer;
}

/*****************************************************************************/

static char *load_template(char *ext)
/*
	load the template file with extension 'ext' into a dynamically allocated
	buffer

	returns a pointer to the buffer on success, or NULL on failure
 */
{
	char	template_path[PATH_MAX];

	/* generate the fully qualified pathname for the template file */
	strcpy(template_path, templates_dir);
	strcat(template_path, TEMPLATENAME);	/* template name minus extension */

	/* add extension to pathname */
	strcat(template_path, ext);

	return load_file(template_path);
}

/*****************************************************************************/

static int load_templates()
/*
	load each of the template files into memory

	returns 0 on success, or -1
*/
{
	/* load the three template files, README.port, README.category and
		README.top
	 */
	if ((template_buf_port = load_template("port")) == NULL)
		return -1;
	if ((template_buf_category = load_template("category")) == NULL)
		return -1;
	if ((template_buf_top = load_template("top")) == NULL)
		return -1;

	return 0;
}

/******************************************************************************/

static int validate_path(const char *path)
/*
	verify that path exists and is a directory

	returns 0 on success, or -1
 */
{
	struct	stat sb;				/* information returned from stat() */

	/* call stat to get info for path */
	if (stat(path, &sb) == -1)
	{
		perror(path);
		return -1;
	}

	/* make sure it's a directory */
	if (!S_ISDIR(sb.st_mode))
	{
		fprintf(stderr, "%s is not a directory\n", path);
		return -1;
	}

	return 0;
}

/******************************************************************************/

char *resolve_path(const char *path)
/*
	Resolve the pathname pointed to by 'path' to a fully qualified pathname

	As a convenience, let any relative path be relative to ports_dir,
	unless the user selected the "-n" option, in which case path will be
	relative to the current working directory

	Returns a pointer to the fully qualified pathname returned by realpath(3),
	or NULL on error

	Uses realpath(3) to do the grunt work

	Caller should free() the returned pointer after it is no longer needed
*/
{
	char *real_path;				/* result from realpath(3) */
	char pathbuf[PATH_MAX];			/* temporary buffer */
	char *ptr;

	if (path == NULL)				/* should never be getting called with */
	{								/* a NULL pointer argument */
		return NULL;
	}

	if (*path == '/')			/* absolute path, use "as is" */
	{
		strcpy(pathbuf, path);
	}
	else if (*path == '.')		/* pathnames beginning with "." or ".." are
									always relative to the current working
									directory, regardless of the setting of the
									normal_paths flag
								 */
	{
		if (getcwd(pathbuf, PATH_MAX) == NULL)
		{
			perror("getcwd");
			return NULL;
		}
		else
		{
			/* append path to current working directory in pathbuf */
			strcat(pathbuf, "/");
			strcat(pathbuf, path);
		}
	}
	else							/* relative path */
	{
		if (!normal_paths)			/* append path to ports_dir */
		{
			strcpy(pathbuf, ports_dir_plus);
			strcat(pathbuf, path);
		}
		else					/* append path to current working directory */
		{
			if (getcwd(pathbuf, PATH_MAX) == NULL)
			{
				perror("getcwd");
				return NULL;
			}
			else
			{
				/* append path to current working directory in pathbuf */
				strcat(pathbuf, "/");
				strcat(pathbuf, path);
			}
		}
	}

	/* Now run what we have through realpath() to "smooth it out"

		Note: paths returned by realpath will *never* have a trailing slash
	 */

	if ((real_path = realpath(pathbuf, NULL)) == NULL)
	{
		perror(pathbuf);
		return NULL;
	}

	return real_path;
}

/*****************************************************************************/

int set_ports_dir(const char *path)
/*
	set the ports_dir variable, a character array containing the name of the
	top-level ports directory

	set ports_dir to 'path' if not NULL, after running some validity checks
	on it first

	if 'path' is NULL, set ports_dir to the current setting of the PORTSDIR
	environment variable, if defined (again, after running some validity
	checks on it), or the default of PORTSTOP ("/usr/ports")

	set ports_dir_plus to ports_dir plus a trailing slash

	returns 0 on success, -1 on error
 */
{
	char   *portsdir_env;	/* current setting of PORTSDIR */
	char   *ptr;			/* general-purpose pointer */

	/*
	 *	First case: (path != NULL), meaning user selected the "-p pathname"
	 *	command line option, which will override any previous setting of the
	 *	ports_dir variable, including the PORTSDIR environment variable
	 */
	if (path != NULL)
	{
		if (strlen(ports_dir))
		{
			/* advise user of the change of path */
			fprintf(stderr, "Changing top-level directory from \"%s\" to \"%s\"\n",
				ports_dir, path);
		}

		if ((ptr = resolve_path(path)) == NULL)
		{
			fprintf(stderr, "Invalid pathname argument: \"%s\"\n", path);
			return -1;
		}

		if (validate_path(ptr) != 0)
		{
			return -1;
		}
		else strcpy(ports_dir, ptr);
	}
	else
	/*
	 *	The other case: (path == NULL), meaning we were called from
	 *	initialize() with a NULL path
	 */
	if ((portsdir_env = getenv("PORTSDIR")) != NULL)
	{
		if ((ptr = resolve_path(portsdir_env)) == NULL)
		{
			fprintf(stderr,
					"Invalid pathname in PORTSDIR environment variable: \"%s\"\n",
					portsdir_env);
			return -1;
		}

		if (validate_path(ptr) != 0)
		{
			return -1;
		}
		else
		{
			strcpy(ports_dir, ptr);
			if (!quiet)
			{
				fprintf(stderr, "Using PORTSDIR (%s) as top-level ports directory\n",
						ports_dir);
			}
		}
	}
	else strcpy(ports_dir, PORTSTOP);	/* use default top */

	/* now set ports_dir_plus to whatever we got for ports_dir, and add a
		trailing slash
	 */
	strcpy(ports_dir_plus, ports_dir);
	strcat(ports_dir_plus, "/");

	return 0;
}

/*****************************************************************************/

int set_templates_dir(char *path)
/*
	set the templates_dir variable

	returns 0 on success, -1 on error
 */
{
	char *ptr;

	if (path != NULL)	/* called from main() by the "-t" option */
	{
		if ((ptr = resolve_path(path)) == NULL)
		{
			fprintf(stderr, "Invalid pathname argument: \"%s\"\n", path);
			return -1;
		}

		if (validate_path(ptr) != 0)
		{
			return -1;
		}
		/* else */
		strcpy(templates_dir, ptr);
	}
	else	/* called from initialize() with NULL path argument */
	{
		/* construct the path to the default Templates directory,
			respecting the PREFIX where the program was originally
			installed
		 */
		strcpy(templates_dir, PREFIX);
		strcat(templates_dir, TEMPLATESDIR);

		if (validate_path(templates_dir) != 0)	/* something's amiss(ing) */
		{
			return -1;
		}
	}

	return 0;
}

/*****************************************************************************/

int initialize()
/*
	Initialize the program by performing the following series of operations:

	Register a cleanup routine to be called automatically when the program
	exits, to ensure that we free up whatever memory we obtain via malloc()

	Unless it was already set with the "-p pathname" command line option,
	set the ports_dir variable to the top-level ports directory, either
	from the environment variable PORTSDIR (if set), or the #defined value of
	PORTSTOP ("/usr/ports/").

	Unless it was already set with the "-t pathname" command line option,
	set the templates_dir variable to the default templates directory
	(${PREFIX}/share/mkreadmes/Templates).

	Generate a list of categories from the top-level Makefile.

	Load the index file	into a dynamically allocated buffer.  The
	load_index_file() routine will also generate a list of pointers to the
	start of each category's section within the index buffer, to optimize
	ports searches later

	Load each of the three template files into their own dynamically allocated
	buffers.

	Returns 0 on success, or -1 on error
*/
{
	/* register a cleanup routine with atexit(3) to automatically
		free whatever memory we allocated when the program exits
	 */
	if (atexit(cleanup) == -1)
	{
		perror("atexit() register of cleanup routine failed");
		return -1;
	}

	/* set the global variables ports_dir and ports_dir_plus,
		if not already set by the "-p" option
	 */
	if (strlen(ports_dir) == 0)
	{
		if (set_ports_dir(NULL) == -1)
		{
			return -1;
		}
	}

	/* set the global variable templates_dir,
		if not already set by the "-t" option
	 */
	if (strlen(templates_dir) == 0)
	{
		if (set_templates_dir(NULL) == -1)
		{
			return -1;
		}
	}

	/* generate a list of categories from the top-level Makefile */
	if (gen_categories_list() != 0)
		return -1;

	/* load the index file into memory */
	if (load_index() == -1)
	{
		fprintf(stderr, "Unable to load the index file!\n");
		return -1;
	}

	/* load all of the template files */
	return load_templates();
}

