/*
 *	File: mkreadmes.c
 *
 *	Contains three routines that drive the body of the work, one for
 *	each level of readmes -- the top level, the category level, and
 *	the individual port level -- plus a few small auxiliary functions
 *
 *	main() will call only one of the make_readme_* routines, based on whether
 *	the pathname given as a command line argument (or the top-level ports
 *	directory, if none is provided) is determined to be at the top, category,
 *	or individual port level
 *
 *	$Id: mkreadmes.c,v 1.99 2012/03/07 16:00:00 conrads Exp $
 *
 *****************************************************************************/

#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "mkreadmes.h"

/* placeholder names used in templates */

static char *placeholder[] = {
	"%%PKG%%",
	"%%PORT%%",
	"%%PREFIX%%",
	"%%COMMENT%%",
	"%%DESCR%%",
	"%%EMAIL%%",
	"%%CATEGORY%%",
	"%%BUILD_DEPENDS%%",
	"%%RUN_DEPENDS%%",
	"%%WEBSITE%%",
	"%%TOP%%",
	"%%SUBDIR%%",
	NULL			/* end of list marker */
};

static	char	category_comment[LINE_MAX];	/* comment from a category's
												Makefile
											 */
/* pointers to dynamically allocated buffers for the template files */
char	*template_buf_top = NULL;
char	*template_buf_category = NULL;
char	*template_buf_port = NULL;

extern	char	ports_dir[PATH_MAX];		/* top-level ports directory */
extern	char	ports_dir_plus[PATH_MAX];	/* ports_dir plus trailing slash */

extern	char	categories[];		/* list of categories found in the top-level
										Makefile
									 */
extern	int		num_categories;		/* number of categories found */
extern	char	*category;			/* pointer to current category */

extern	char	**index_categories;	/* pointer to an array of pointers to the
										first line of each category's section
										within the index buffer
									 */
extern	char	*index_field[];	/* pointers to current port's index fields */

extern	char	*next_index;	/* pointer to the next index line to search */
extern	char	*next_cat;	/* pointer to first index line of next category */

extern	int		quiet;			/* silence normal output */
extern	int		verbose;		/* increase the amount of output */

/* prototypes for routines in other source files */

extern	int		search_index(const char *path);					   /* index.c */
extern	char   *skip_leading(const char *str, const char *pattern);	/* init.c */

/*****************************************************************************/

static char *HTMLify(char *ptr, char *str )
/*
	Copy 'str' to 'ptr', replacing any occurences of '&', '<' or '>'
	with their HTML	representations

	Returns a pointer to the location in 'ptr' one past the last
	character copied

	Uses the addstr() macro defined in mkreadmes.h
 */
{
	while (*str)
	{
		switch(*str)
		{
			case '&':
			{
				addstr(ptr, "&amp;");
				++str;
				continue;
			}
			case '<':
			{
				addstr(ptr, "&lt;");
				++str;
				continue;
			}
			case '>':
			{
				addstr(ptr, "&gt;");
				++str;
				continue;
			}
			default:
				*ptr++ = *str++;
		}
	}

	return ptr;
}

/*****************************************************************************/

static int match_placeholder(const char *ptr)
/*
	Try to find a matching "%%VARIABLE%%" within the placeholder array
	for the string pointed to by 'ptr'

	On entry to this routine, 'ptr' will be pointing to the first "%" in the
	possible symbol name we're looking for, so all we have to do is see if the
	next character is also a "%" and then scan the placeholder array for a match

	Returns a positive integer representing the variable's index in the
	placeholder array, or -1 if no match is found
 */
{
	if (ptr[1] == '%')
	{
		int i = 0;

		/* We've hit a double "%%", now find its match in the placeholder
			array, and return its index within the placeholder array
		 */

		while (placeholder[i])
		{
			if (strncmp(ptr, placeholder[i], strlen(placeholder[i])) == 0)
				return i;
			++i;
		}
	}

	/* No match found */
	return -1;
}

/*****************************************************************************/

static int category_index(char *category)
/*
	scan the list of category names to get an index to use with the
	index_categories[] array to get the pointer to the first index line of a
	category in the index buffer

	returns the index at which the category was found in the list, or
	num_categories if not found
*/
{
	char   *categories_ptr = categories;
	int		i = 0;

	while (*categories_ptr)
	{
		if (strcmp(category, categories_ptr) == 0)
			return i;									/* found a match */
		/* else */
		categories_ptr += strlen(categories_ptr) + 1;	/* bump pointer and */
		++i;											/* index */
	}

	return i;	/* (i == num_categories) = not found */
}

/*****************************************************************************/

static int make_readme_port(const char *path, int make_level)
/*
	Make the README.html file for a single port

	On entry, category will already be pointing either to the "category/port"
	part of the port's pathname (if make_level == PORT) or just the category
	part (if make_level == CATEGORY or TOP)

	Returns 0 on success, or 1 on failure

	Uses the addstr() macro defined in mkreadmes.h
*/
{
	char	readme_buf[MAXREADME_PORT];	/* output buffer */
	char   *readme_ptr = readme_buf;	/* pointer into output buffer */
	char	readme_path[PATH_MAX];		/* output pathname */
	int		readme;						/* file descriptor */

	char   *template_ptr = template_buf_port;	/* pointer into template
													buffer
												 */
	char   *catport;					/* category/port portion of path */
	char	catonly[MAXCATNAME];		/* just the category name */
	int		catlen;						/* length of category name */

	int		i;					/* index into the placeholder array where a
									template match was found
								 */
	int		cat_index;	/* index into the categories[] and index_categories[]
							arrays for this port's category (can't initialize
							here as we do in make_readme_category() below, not
							until we make certain that 'category' has been set
							properly
						 */

	/* save a pointer to the category/port portion of port's path */
	catport = skip_leading(path, ports_dir_plus);

	if (make_level == PORT)	/* need to set category, next_index and next_cat */
	{
		/* copy just the category to catonly */
		catlen = strcspn(catport, "/");
		strncpy(catonly, catport, catlen);
		catonly[catlen] = '\0';
		category = catonly;			/* reset category (set earlier by
										get_make_level()) to point to the
										string we just copied
									 */
	/* set next_index to point to the first line of this port's category's
		section in the index buffer, so we won't waste any time searching in
		any of the preceding categories
	 */
		next_index = index_categories[cat_index = category_index(category)];

		if (next_index == NULL)	/* category not found :-( */
		{
			fprintf(stderr, "No such category: %s\n", category);
			return 1;
		}

		/* set next_cat to point to the first index line of the next category,
			to use as the search cutoff point
		 */
		next_cat = index_categories[cat_index + 1];
	}

	/* No "else" needed here.  If make_level > PORT, then category already
		points to the category alone, and next_index and next_cat have been set
		by make_readme_category().
	 */

	/* Find the port's entry in the index file, saving pointers to its index
		information in the index_field[] array
	 */
	if (!search_index(path))
	{
		fprintf(stderr, "Port %s not found in INDEX\n", catport);
		return 1;
	}

	/* For efficiency's sake, don't output the following message
		unless this one port is all we're building, or the verbose
		flag is on
	 */
	if (verbose || ((make_level == PORT) && !quiet))
		printf("Making README.html for port %s\n", catport);

	/* Open the port's README.html file for writing */

	strcpy(readme_path, path);			/* port's path in the ports tree */
	strcat(readme_path, READMENAME);	/* append "/README.html" */

	if ((readme = open(readme_path, O_WRONLY|O_CREAT|O_TRUNC|O_DIRECT, MODE))
				== -1)
	{
		perror(readme_path);
		return 1;
	}

	/* Copy the template buffer to the readme buffer,
		performing any substitutions needed
	 */

	while (*template_ptr)
	{
		/* just copy from template buffer to readme buffer
			unless we hit a '%'
		 */
		if ((*template_ptr == '%') &&
			((i = match_placeholder(template_ptr)) >= 0))
		{
			/* a few fields require special handling:

				for the PORT field, use the "category/port" string we saved
				earlier in the variable catport[]

				for the COMMENT field, convert any '&', '<' or '>' characters
				into their HTML representations

				for the DESCR field, adjust the path prefix to the realpath
				of the top of our ports tree, which may differ from the
				canonical "/usr/ports" used in the index field

				for BUILD_DEPENDS and RUN_DEPENDS, don't output anything
				if these fields are empty in the INDEX file, otherwise,
				precede them with the introductory text #defined in
				BUILD_DEPS or RUN_DEPS

				for WEBSITE, output the lead-in text #defined in WWW_BEFORE,
				then the field contents, and follow it up with the text
				#defined in WWW_AFTER

			 */
			switch (i)
			{
				case PKG:
					addstr(readme_ptr, index_field[i]);
					break;
				case PORT:
					addstr(readme_ptr, catport);
					break;
				case COMMENT:
					/* the port's comment may have some bad HTML chars */
					readme_ptr = HTMLify(readme_ptr, index_field[COMMENT]);
					break;
				case DESCR:
					addstr(readme_ptr, ports_dir);
					addstr(readme_ptr, skip_leading(index_field[i], PORTSTOP));
					break;
				case EMAIL:
					addstr(readme_ptr, index_field[i]);
					break;
				case BUILD_DEPENDS:
					if (strlen(index_field[i]))
					{
						addstr(readme_ptr, BUILD_DEPS);
						addstr(readme_ptr, index_field[i]);
					}
					break;
				case RUN_DEPENDS:
					if (strlen(index_field[i]))
					{
						addstr(readme_ptr, RUN_DEPS);
						addstr(readme_ptr, index_field[i]);
					}
					break;
				case WEBSITE:
					if (strlen(index_field[i]))
					{
						addstr(readme_ptr, WWW_BEFORE);
						addstr(readme_ptr, index_field[i]);
						addstr(readme_ptr, WWW_AFTER);
					}
					break;
				case TOP:
					addstr(readme_ptr, ports_dir);
					break;
			}
			/* bump template pointer past placeholder */
			template_ptr += strlen(placeholder[i]);
		}
		else *readme_ptr++ = *template_ptr++;
	}

	/* write readme_buf to output file and close */

	if (write(readme, readme_buf, (readme_ptr - readme_buf)) == -1)
	{
		perror(readme_path);
		close(readme);
		return 1;
	}

	close(readme);

	return 0;
}

/*****************************************************************************/

static int make_readme_category(const char *path, int make_level)
/*
	Make the README.html file for a single category

	Calls make_readme_port() to build the README.html file for each
	port within the category

	Returns 0 on success, or 1 on failure

	Uses the addstr() macro defined in mkreadmes.h
*/
{
	char	readme_buf[MAXREADME_CAT];	/* output buffer */
	char   *readme_ptr = readme_buf;	/* pointer into output buffer */
	char	readme_path[PATH_MAX];		/* output pathname */
	int		readme;						/* file descriptor */

	char	Makefile_buf[LINE_MAX];		/* input buffer */
	char	Makefile_path[PATH_MAX];	/* input pathname */
	FILE   *Makefile;					/* input stream for category's
											Makefile
										 */
	char   *comment_ptr = Makefile_buf;	/* for locating the COMMENT line */
	char   *template_ptr = template_buf_category;	/* pointer into template
														buffer
													 */
	char	ports[MAXPORTSLIST];		/* list of ports found in a category's
											Makefile
										 */
	int		num_ports = 0;		/* number of ports in a category */
	char   *ports_ptr = ports;	/* pointer into ports list */
	char   *port;				/* pointer to port in current input buffer */
	char	port_path[PATH_MAX];	/* full path to port in ports tree */

	int		i;					/* index into the placeholder array where a
									template match was found
								 */

/* index into the categories[] and index_categories[] arrays
	for this category
 */
	int		cat_index = category_index(category);

	/* set next_index to point to the first line of this category's section
		in the index buffer, so we won't waste any time searching in any of
		the preceding categories
	 */
	if ((next_index = index_categories[cat_index]) == NULL)	/* not found */
	{
		fprintf(stderr, "No such category: %s\n", category);
		return 1;
	}

	/* set next_cat to point to the first index line of the next category,
		to use as the search cutoff point
	 */
	next_cat = index_categories[cat_index + 1];

	/* Open the category's Makefile for reading */

	strcpy(Makefile_path, path);
	strcat(Makefile_path, "/");
	strcat(Makefile_path, MAKEFILENAME);

	if ((Makefile = fopen(Makefile_path, "r")) == NULL)
	{
		perror(Makefile_path);
		return 1;
	}

	/* First, find the COMMENT line, saving it externally in category_comment[]
		in case we were called from make_readme_top(), which will need the
		comment for its own README.html file
	 */

	while ((fgets(Makefile_buf, LINE_MAX, Makefile) != NULL))
	{
		if ((comment_ptr = skip_leading(Makefile_buf, "    COMMENT = "))
			!= Makefile_buf)	/* found it */
		{
			/* don't want the newline */
			comment_ptr[strlen(comment_ptr) - 1] = '\0';
			strcpy(category_comment, comment_ptr);	/* save the COMMENT */
			break;
		}
	}

	/* Got the comment, continue reading to get all the SUBDIRs */

	while ((fgets(Makefile_buf, LINE_MAX, Makefile) != NULL))
	{
		if ((port = skip_leading(Makefile_buf, "    SUBDIR += ")) !=
				Makefile_buf)	/* found a SUBDIR line */
		{
			/* don't want the newline */
			port[strlen(port) - 1] = '\0';
			ports_ptr = stpcpy(ports_ptr, port) + 1;	/* add to ports list */
			++num_ports;
		}
	}

	/* add an extra nul character to mark the end of the ports list */
	*ports_ptr = '\0';

	fclose(Makefile);

	/* Done collecting the data, now we can start generating
		the README.html file
	 */
	if (!quiet)
		printf("Making README.html for category %s (%d ports)\n",
			category, num_ports);

	/* Open the category's README.html file for writing */

	strcpy(readme_path, path);
	strcat(readme_path, "/");
	strcat(readme_path, READMENAME);

	if ((readme = open(readme_path, O_WRONLY|O_CREAT|O_TRUNC|O_DIRECT, MODE))
				== -1)
	{
		perror(readme_path);
		return 1;
	}

	/* Copy the template buffer to the readme buffer,
		performing any substitutions needed
	 */

	while (*template_ptr)
	{
		/* just copy from template buffer to readme buffer
			unless we hit a '%'
		 */
		if ((*template_ptr == '%') &&
			((i = match_placeholder(template_ptr)) >= 0))
		{
			/* SUBDIR requires special handling, calling make_readme_port()
				for each port in the category to build the port's README.html
				file and to obtain its package name and comment for the
				category README.html file

				just as in make_readme_port(), for the port's COMMENT field,
				convert any '&', '<' or '>' characters into their HTML
				representations
			 */
			switch (i)
			{
				case COMMENT:	/* the category's comment, not a port's */
					addstr(readme_ptr, category_comment);
					/* add the number of ports in category to end of comment */
					readme_ptr += sprintf(readme_ptr, " (%d ports)", num_ports);
					break;
				case CATEGORY:
					addstr(readme_ptr, category);
					break;
				case SUBDIR:
					ports_ptr = ports;

					while (*ports_ptr)
					{
						port = ports_ptr;	/* port's base name */

						/* generate the full path to the port */
						strcpy(port_path, path);	/* category's path */
						strcat(port_path, "/");
						strcat(port_path, port);	/* append port's name */

						/* make the port's README.html */
						make_readme_port(port_path, make_level);

						/* now we can use the information left over in the
							index_field[] array by make_readme_port() */
						addstr(readme_ptr, "<a href=\"");
						addstr(readme_ptr, port);
						addstr(readme_ptr, "/README.html\">");
						addstr(readme_ptr, index_field[PKG]);
						addstr(readme_ptr, "</a>: ");

						/* the port's comment may have some bad HTML chars */
						readme_ptr = HTMLify(readme_ptr, index_field[COMMENT]);
						strcpy(readme_ptr++, "\n");

						/* bump ports_ptr to next item in ports array */
						ports_ptr += strlen(port) + 1;
					}
			}
			/* bump template pointer past placeholder */
			template_ptr += strlen(placeholder[i]);
		}
		else *readme_ptr++ = *template_ptr++;
	}

	/* write readme_buf to output file and close */

	if (write(readme, readme_buf, (readme_ptr - readme_buf)) == -1)
	{
		perror(readme_path);
		close(readme);
		return 1;
	}

	close(readme);

	return 0;
}

/*****************************************************************************/

static int make_readme_top(const char *path, int make_level)
/*
	Make the top-level README.html file

	Calls make_readme_category() to build the README.html file for each
	category in the categories list created earlier from the top-level Makefile

	Returns 0 on success, or 1 on failure

	Uses the addstr() macro defined in mkreadmes.h
*/
{
	char	readme_buf[MAXREADME_TOP];	/* output buffer */
	char   *readme_ptr = readme_buf;	/* pointer into output buffer */
	char	readme_path[PATH_MAX];		/* output pathname */
	int		readme;						/* file descriptor */

	char   *template_ptr = template_buf_top;	/* pointer into template buffer
												 */
	char   *categories_ptr = categories;	/* pointer into categories list */

	char	category_path[PATH_MAX];	/* fully qualified path to current
											category
										 */
	int		i;						/* index into the placeholder array where
										a template match was found
									 */
	if (!quiet)
		printf("Making top-level README.html in %s (%d categories)\n",
			ports_dir, num_categories);

	/* Open the top-level README.html file for writing */

	strcpy(readme_path, path);			/* fully qualified top-level path */
	strcat(readme_path, READMENAME);	/* append "/README.html" */

	if ((readme = open(readme_path, O_WRONLY|O_CREAT|O_TRUNC|O_DIRECT, MODE))
				== -1)
	{
		perror(readme_path);
		return 1;
	}

	/* Copy the template buffer to the readme buffer,
		performing any substitutions needed
	 */

	while (*template_ptr)
	{
		/* just copy from template buffer to readme buffer
			unless we hit a '%'
		 */
		if ((*template_ptr == '%') &&
			((i = match_placeholder(template_ptr)) >= 0))
		{
			/* %%SUBDIR%% is the only placeholder used in the top-level
				template, and will cause this routine to call
				make_readme_category() for each category found in the top-level
				Makefile (already obtained and saved earlier in the categories[]
				list), which in turn will call make_readme_port() for each port
				in the category's own Makefile.

				On return from make_readme_category() the category's COMMENT
				will be stored in category_comment[] for use in the top-level
				README.html file.
			 */
			if (i == SUBDIR)
			{
				categories_ptr = categories;

				while (*categories_ptr)
				{
					category = categories_ptr;	/* base name of category */

					/* generate the full path to the category */
					strcpy(category_path, ports_dir_plus);
					strcat(category_path, category);

					/* make the category's README.html, saving its
						COMMENT in category_comment[] */
					make_readme_category(category_path, make_level);

					/* now we have all the information we need
						for this category
					 */
					addstr(readme_ptr, "<a href=\"");
					addstr(readme_ptr, category);
					addstr(readme_ptr, "/README.html\">");
					addstr(readme_ptr, category);
					addstr(readme_ptr, "</a>: ");
					addstr(readme_ptr, category_comment);
					strcpy(readme_ptr++, "\n");

					/* bump pointer to next item in categories array */
					categories_ptr += strlen(category) + 1;
				}
			}
			/* bump template pointer past placeholder */
			template_ptr += strlen(placeholder[i]);
		}
		else *readme_ptr++ = *template_ptr++;
	}

	/* write readme_buf to output file and close */

	if (write(readme, readme_buf, (readme_ptr - readme_buf)) == -1)
	{
		perror(readme_path);
		close(readme);
		return 1;
	}

	close(readme);

	return 0;
}

/*****************************************************************************/

int make_readme(const char *path, int make_level)
/*
	wrapper for the actual make_readme_* routines, called from main()

	returns the return code of whichever routine it calls

	(the default case should never happen, and is provided merely to avoid
	a warning from the compiler)
 */
{
	switch (make_level)
	{
		case TOP:		return make_readme_top(path, make_level);
		case CATEGORY:	return make_readme_category(path, make_level);
		case PORT:		return make_readme_port(path, make_level);
		default:		return -1;
	}
}

